
/* ------ oscils, lphasor, and tablexkt by Istvan Varga (Jan 5 2002) ------ */

#include "cs.h"
#include <math.h>
#define CSOUND_OSCILS_C 1
#include "oscils.h"

/* ------------- set up fast sine generator ------------- */
/* Input args:                                            */
/*   a: amplitude                                         */
/*   f: frequency (-PI - PI)                              */
/*   p: initial phase (0 - PI/2)                          */
/* Output args:                                           */
/*  *x: first output sample                               */
/*  *c, *v: coefficients for calculating next sample as   */
/*          shown below:                                  */
/*            v = v + c * x                               */
/*            x = x + v                                   */
/*          These values are calculated by:               */
/*            x = y[0]                                    */
/*            c = 2.0 * cos(f) - 2.0                      */
/*            v = y[1] - (c + 1.0) * y[0]                 */
/*          where y[0], and y[1] are the first, and       */
/*          second sample of the sine wave to be          */
/*          generated, respectively.                      */
/* -------- written by Istvan Varga, Jan 28 2002 -------- */

static void init_sine_gen (double a, double f, double p,
                           double *x, double *c, double *v)
{
    double  y0, y1;                 /* these should be doubles */

    y0 = sin (p);
    y1 = sin (p + f);
    *x = y0;
    *c = 2.0 * cos (f) - 2.0;
    *v = y1 - *c * y0 - y0;
    /* amp. scale */
    *x *= a; *v *= a;
}


/* -------- oscils set-up -------- */

void oscils_set (OSCILS *p)
{
    int     iflg;

    iflg = (int) (*(p->iflg) + FL(0.5)) & 0x07;   /* check flags */
    if (iflg & 1) return;           /* skip init, nothing to do */
    p->use_double = (iflg & 2 ? 1 : 0);           /* use doubles internally */
    init_sine_gen ((double) *(p->iamp), (double) (*(p->icps) * tpidsr),
                   (double) (*(p->iphs) * TWOPI_F),
                   &(p->xd), &(p->cd), &(p->vd));
    if (!(p->use_double)) {
      p->x = (MYFLT) p->xd;       /* use floats */
      p->c = (MYFLT) p->cd;
      p->v = (MYFLT) p->vd;
    }
}

/* -------- oscils performance -------- */

void oscils (OSCILS *p)
{
    int     nn;
    MYFLT   *ar, x, c, v;
    double  xd, cd, vd;

    /* copy object data to local variables */
    ar = p->ar;

    nn = ksmps;
    if (p->use_double) {            /* use doubles */
      xd = p->xd; cd = p->cd; vd = p->vd;
      while (nn--) {
        *(ar++) = (MYFLT) xd;
        vd += cd * xd;
        xd += vd;
      }
      p->xd = xd; p->vd = vd;
    }
    else {                          /* use floats */
      x = p->x; c = p->c; v = p->v;
      while (nn--) {
        *(ar++) = x;
        v += c * x;
        x += v;
      }
      p->x = x; p->v = v;
    }
}

/* -------- lphasor set-up -------- */

void lphasor_set (LPHASOR *p)
{
    if (*(p->istor) != FL(0.0)) return;               /* nothing to do */

    p->phs = (double) *(p->istrt);                          /* start phase */
    p->lps = (double) *(p->ilps);                           /* loop start */
    p->lpe = (double) *(p->ilpe);                           /* loop end */
    p->loop_mode = (int) (*(p->imode) + FL(0.5)) & 0x03;    /* loop mode */
    if (p->lpe <= p->lps) p->loop_mode = 0;                 /* disable loop */
    p->dir = 1;                                             /* direction */
}

/* -------- lphasor performance -------- */

void lphasor (LPHASOR *p)
{
    int     nn, loop_mode, dir;
    MYFLT   *ar, *xtrns;
    double  trns, phs, lps, lpe, lpt;

    /* copy object data to local variables */
    ar = p->ar; xtrns = p->xtrns;
    phs = p->phs; lps = p->lps; lpe = p->lpe;
    lpt = lpe - lps;
    loop_mode = p->loop_mode;
    trns = (double) *xtrns;

    nn = ksmps;
    while (nn--) {
      if (XINARG1) trns = (double) *(xtrns++);
      *(ar++) = (MYFLT) phs;
      phs += (p->dir ? trns : -trns);
      if (loop_mode) {
        dir = (trns < 0.0 ? !(p->dir) : p->dir);
        if (dir && (phs >= lpe)) {
          phs += lpt * (double) ((int) (lps - phs) / lpt);
          if (loop_mode & 2) {
            phs = lps + lpe - phs;  /* reverse direction */
            p->dir = !(p->dir);
          }
        }
        else if (!dir && (phs <= lps)) {
          phs += lpt * (double) ((int) (lpe - phs) / lpt);
          if (loop_mode & 1) {
            phs = lps + lpe - phs;  /* reverse direction */
            p->dir = !(p->dir);
          }
        }
      }
    }
    /* store phase */
    p->phs = phs;
}

/* -------- tablexkt set-up -------- */

void tablexkt_set (TABLEXKT *p)
{
    p->wsize = (int) (*(p->iwsize) + 0.5);                  /* window size */
    if (p->wsize < 3) {
      p->wsize = 2;
    }
    else {
      p->wsize = ((p->wsize + 2) >> 2) << 2;        /* round to nearest */
      if (p->wsize > 1024) p->wsize = 1024;         /* integer multiply of 4 */
    }
    /* constant for window calculation */
    p->win_fact = (MYFLT) ((1.0 - pow ((double) p->wsize * 0.85172, -0.89624))
                           / (double) ((p->wsize * p->wsize) >> 2));

    p->ndx_scl = (*(p->ixmode) == FL(0.0) ? 0 : 1);         /* index mode */
    p->wrap_ndx = (*(p->iwrap) == FL(0.0) ? 0 : 1);         /* wrap index */
    /* use raw index values without scale / offset */
    if ((*(p->ixoff) != FL(0.0)) || p->ndx_scl) p->raw_ndx = 0;
    else p->raw_ndx = 1;
}

/* -------- tablexkt opcode -------- */

void tablexkt (TABLEXKT *p)
{
    int     i, nn, wsize, wsized2, wrap_ndx, warp;
    double  ndx, d, x, c, v, flen_d, onedpi_d, pidwarp_d;
    long    ndx_i, flen;
    MYFLT   *ar, *xndx, ndx_f, a0, a1, a2, a3, v0, v1, v2, v3, *ftable;
    MYFLT   onedwarp, win_fact;
    FUNC    *ftp;

    /* window size */
    wsize = p->wsize;
    if ((wsize < 2) || (wsize > 1024)) {
      perferror (Str(X_1714,"tablexkt: not initialised")); return;
    }
    wsized2 = wsize >> 1;

    /* check ftable */
    if ((ftp = ftfindp (p->kfn)) == NULL) return;     /* invalid table */
    if ((ftable = ftp->ftable) == NULL) return;
    flen = ftp->flen;               /* table length */
    flen_d = (double) flen;

    /* copy object data to local variables */
    ar = p->ar;
    xndx = p->xndx;
    wrap_ndx = p->wrap_ndx;
    if ((wsize > 4) && (*(p->kwarp) > FL(1.001))) {
      warp = 1;                     /* enable warp */
      onedwarp = FL(1.0) / *(p->kwarp);
      pidwarp_d = PI / (double) *(p->kwarp);
      /* correct window for kwarp */
      x = v = (double) wsized2; x *= x; x = 1.0 / x;
      v *= (double) onedwarp; v -= (double) ((int) v) + 0.5; v *= 4.0 * v;
      win_fact = (MYFLT) (((double) p->win_fact - x) * v + x);
    }
    else {
      warp = 0; onedwarp = FL(0.0); pidwarp_d = 0.0;
      win_fact = p->win_fact;
    }
    onedpi_d = 1.0 / PI;

    nn = ksmps;
    while (nn--) {
      ndx = (double) *xndx;
      if (XINARG1) xndx++;
      /* calculate table index */
      if (!(p->raw_ndx)) {
        ndx += (double) *(p->ixoff);
        if (p->ndx_scl) ndx *= flen_d;
      }
      /* integer and fractional part of table index */
      ndx_i = (long) ndx;
      ndx_f = (MYFLT) (ndx - (double) ndx_i);
      if (ndx_f < FL(0.0)) {
        ndx_f++; ndx_i--;
      }
      /* wrap or limit to allowed range */
      if (wrap_ndx) {
        while (ndx_i >= flen) ndx_i -= flen;
        while (ndx_i < 0) ndx_i += flen;
      }
      else {                        /* limit */
        if (ndx_i < 0) {
          ndx_i = 0; ndx_f = FL(0.0);
        }
        else if (ndx_i >= flen) {
          ndx_i = flen; ndx_f = FL(0.0);
        }
      }
      switch (wsize) {
        case 2:                     /* ---- linear interpolation ---- */
          *ar = ftable[ndx_i];
          if (++ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
          *ar += (ftable[ndx_i] - *ar) * ndx_f;
          break;
        case 4:                     /* ---- cubic interpolation ---- */
          /* sample  0 */
          v1 = ftable[ndx_i];
          /* sample -1 */
          if (ndx_i) {
            v0 = ftable[ndx_i - 1L];
          } else {
            v0 = ftable[(wrap_ndx ? flen - 1L : 0L)];
          }
          /* sample +1 */
          if (++ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
          v2 = ftable[ndx_i];
          /* sample +2 */
          if (++ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
          v3 = ftable[ndx_i];
          a3 = ndx_f * ndx_f; a3--; a3 *= FL(0.1666666667);
          a2 = ndx_f; a2++; a0 = (a2 *= FL(0.5)); a0--;
          a1 = FL(3.0) * a3; a2 -= a1; a0 -= a3; a1 -= ndx_f;
          *ar = (a0 * v0 + a1 * v1 + a2 * v2 + a3 * v3) * ndx_f + v1;
          break;
        default:                    /* ---- sinc interpolation ---- */
          *ar = FL(0.0);        /* clear output */
          ndx = (double) ndx_f;
          ndx_i += (long) (1 - wsized2);
          d = (double) (1 - wsized2) - ndx;
          if (warp) {           /* ---- warp enabled ---- */
            init_sine_gen (onedpi_d, pidwarp_d, pidwarp_d * d, &x, &c, &v);
            /* samples -(window size / 2 - 1) to -1 */
            i = wsized2 - 1;
            do {
              a1 = (MYFLT) d; a1 = FL(1.0) - a1 * a1 * win_fact;
              a1 = a1 * a1 / (MYFLT) d;
              *ar += ftable[(ndx_i < 0L ? (wrap_ndx ? ndx_i + flen : 0L)
                                          : ndx_i)] * (MYFLT) x * a1;
              ndx_i++;
              d++; v += c * x; x += v;
            } while (--i);
            /* sample 0 */
            /* avoid division by zero */
            if (ndx < 0.00003) *ar += onedwarp * ftable[ndx_i];
            else {
              a1 = (MYFLT) d; a1 = FL(1.0) - a1 * a1 * win_fact;
              a1 = a1 * a1 / (MYFLT) d;
              *ar += (MYFLT) x * a1 * ftable[ndx_i];
            }
            d++; v += c * x; x += v;
            if (++ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
            /* sample 1 */
            /* avoid division by zero */
            if (ndx > 0.99997) *ar += onedwarp * ftable[ndx_i];
            else {
              a1 = (MYFLT) d; a1 = FL(1.0) - a1 * a1 * win_fact;
              a1 = a1 * a1 / (MYFLT) d;
              *ar += (MYFLT) x * a1 * ftable[ndx_i];
            }
            d++; v += c * x; x += v;
            if (++ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
            /* samples 2 to (window size / 2) */
            i = wsized2 - 1;
            do {
              a1 = (MYFLT) d; a1 = FL(1.0) - a1 * a1 * win_fact;
              a1 = a1 * a1 / (MYFLT) d;
              *ar += (MYFLT) x * a1 * ftable[ndx_i];
              d++; v += c * x; x += v;
              if (++ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
            } while (--i);
          }
          else {                /* ---- warp disabled ---- */
            /* avoid division by zero */
            if (ndx < 0.00001) {
              ndx_i += (long) (wsized2 - 1);    /* no need to check here */
              *ar = ftable[ndx_i];
            }
            else if (ndx > 0.99999) {
              ndx_i += (long) wsized2;          /* does need range checking */
              if (ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
              *ar = ftable[ndx_i];
            }
            else {
              /* samples -(window size / 2 - 1) to 0 */
              i = wsized2 >> 1;
              do {
                a1 = (MYFLT) d; a1 = FL(1.0) - a1 * a1 * win_fact;
                *ar += ftable[(ndx_i < 0L ? (wrap_ndx ? ndx_i + flen : 0L)
                                            : ndx_i)] * a1 * a1 / (MYFLT) d;
                d++; ndx_i++;
                a1 = (MYFLT) d; a1 = FL(1.0) - a1 * a1 * win_fact;
                *ar -= ftable[(ndx_i < 0L ? (wrap_ndx ? ndx_i + flen : 0L)
                                            : ndx_i)] * a1 * a1 / (MYFLT) d;
                d++; ndx_i++;
              } while (--i);
              /* samples 1 to (window size / 2) */
              i = wsized2 >> 1;
              do {
                a1 = (MYFLT) d; a1 = FL(1.0) - a1 * a1 * win_fact;
                *ar += a1 * a1 * ftable[ndx_i] / (MYFLT) d;
                d++;
                if (++ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
                a1 = (MYFLT) d; a1 = FL(1.0) - a1 * a1 * win_fact;
                *ar -= a1 * a1 * ftable[ndx_i] / (MYFLT) d;
                d++;
                if (++ndx_i >= flen) ndx_i = (wrap_ndx ? ndx_i - flen : flen);
              } while (--i);
              *ar *= (MYFLT) (sin (PI * ndx) / PI);
            }
          }
          break;
      }
      ar++;             /* next sample */
    }
}

/* IV - Dec 20 2002: new opcodes: vincr, vdecr, vsub, vdiv, Inc, Dec, and dec */

typedef struct {
        OPDS    h;
        MYFLT   *r, *a, *b;
} AOP;

#define ARITHM_OP_KK(name,a,op,b)                                            \
void name(AOP *p)                                                            \
{                                                                            \
    *(a) op b;                                                               \
}

#define ARITHM_OP_KA(name,a,op,b)                                            \
void name(AOP *p)                                                            \
{                                                                            \
    *(a) op *(b);                                                            \
}

#define ARITHM_OP_AK(name,a,op,b)                                            \
void name(AOP *p)                                                            \
{                                                                            \
    MYFLT   *x1 = (a);                                                       \
    int     nn = ksmps;                                                      \
    do                                                                       \
      *x1++ op (b);                                                          \
    while (--nn);                                                            \
}

#define ARITHM_OP_AA(name,a,op,b)                                            \
void name(AOP *p)                                                            \
{                                                                            \
    MYFLT   *x1 = (a), *x2 = (b);                                            \
    int     nn = ksmps;                                                      \
    do                                                                       \
      *x1++ op *x2++;                                                        \
    while (--nn);                                                            \
}

/* Inc xsig: xsig = xsig + 1 */

ARITHM_OP_KK(inc_k,p->r,+=,FL(1.0))
ARITHM_OP_AK(inc_a,p->r,+=,FL(1.0))

/* Dec xsig or dec xsig: xsig = xsig - 1 */

ARITHM_OP_KK(dec_k,p->r,-=,FL(1.0))
ARITHM_OP_AK(dec_a,p->r,-=,FL(1.0))

/* shl xsig: xsig = xsig + xsig */

ARITHM_OP_KK(shl_k,p->r,+=,*(p->r))
ARITHM_OP_AA(shl_a,p->r,+=,p->r)

/* sqr xsig: xsig = xsig * xsig */

ARITHM_OP_KK(sqr_k,p->r,*=,*(p->r))
ARITHM_OP_AA(sqr_a,p->r,*=,p->r)

/* vincr xsig1, xsig2: xsig1 = xsig1 + xsig2 */

ARITHM_OP_KK(vincr_kk,p->r,+=,*(p->a))
ARITHM_OP_KA(vincr_ka,p->r,+=,p->a)
ARITHM_OP_AK(vincr_ak,p->r,+=,*(p->a))
ARITHM_OP_AA(vincr_aa,p->r,+=,p->a)

/* vdecr xsig1, xsig2: xsig1 = xsig1 - xsig2 */

ARITHM_OP_KK(vdecr_kk,p->r,-=,*(p->a))
ARITHM_OP_KA(vdecr_ka,p->r,-=,p->a)
ARITHM_OP_AK(vdecr_ak,p->r,-=,*(p->a))
ARITHM_OP_AA(vdecr_aa,p->r,-=,p->a)

/* vmul xsig1, xsig2: xsig1 = xsig1 * xsig2 */

ARITHM_OP_KK(vmul_kk,p->r,*=,*(p->a))
ARITHM_OP_KA(vmul_ka,p->r,*=,p->a)
ARITHM_OP_AK(vmul_ak,p->r,*=,*(p->a))
ARITHM_OP_AA(vmul_aa,p->r,*=,p->a)

/* vdiv xsig1, xsig2: xsig1 = xsig1 / xsig2 */

ARITHM_OP_KK(vdiv_kk,p->r,/=,*(p->a))
ARITHM_OP_KA(vdiv_ka,p->r,/=,p->a)
ARITHM_OP_AK(vdiv_ak,p->r,/=,*(p->a))
ARITHM_OP_AA(vdiv_aa,p->r,/=,p->a)

/* ---- denorm opcode ---- */

static int denorm_seed = 0;

#ifdef MYFLT_IS_FLOAT
#define DENORM_RND      ((MYFLT) ((denorm_seed = (denorm_seed * 15625 + 1)   \
                                                 & 0xFFFF) - 0x8000)         \
                                  * FL(1.0e-24))
#else
#define DENORM_RND      ((MYFLT) ((denorm_seed = (denorm_seed * 15625 + 1)   \
                                                 & 0xFFFF) - 0x8000)         \
                                  * FL(1.0e-60))
#endif

void denorms(DENORMS *p)
{
    MYFLT   r, *ar, **args = p->ar;
    int     n = p->INOCOUNT, nn;

    do {
      r = DENORM_RND;
      ar = *args++;
      nn = ksmps;
      do {
        *ar++ += r;
      } while (--nn);
    } while (--n);
}

/* ---- delayk and vdelayk opcodes ---- */

void delaykset(DELAYK *p)
{
    int npts, mode = (int) (*p->imode + FL(0.5)) & 3;

    if (mode & 1) return;                       /* skip initialisation */
    p->mode = mode;
    /* calculate delay time */
    npts = (int) (*p->idel * ekr + FL(1.5));
    if (npts < 1) {
      initerror("delayk: invalid delay time (must be >= 0)");
      return;
    }
    p->readp = 0; p->npts = npts;
    /* allocate space for delay buffer */
    if (p->aux.auxp == NULL ||
        (npts * (int) sizeof(MYFLT)) > p->aux.size) {
      auxalloc((long) (npts * sizeof(MYFLT)), &p->aux);
    }
    p->init_k = npts - 1;
}

void delayk(DELAYK *p)
{
    MYFLT   *buf = (MYFLT*) p->aux.auxp;

    if (!buf) {
      perferror("delayk: not initialised");
      return;
    }
    buf[p->readp++] = *(p->ksig);           /* write input signal to buffer */
    if (p->readp >= p->npts) p->readp = 0;  /* wrap index */
    if (p->init_k) {
      *(p->ar) = (p->mode & 2 ? *(p->ksig) : FL(0.0));  /* initial delay */
      p->init_k--;
    }
    else
      *(p->ar) = buf[p->readp];             /* read output signal */
}

void vdelaykset(VDELAYK *p)
{
    int npts, mode = (int) (*p->imode + FL(0.5)) & 3;

    if (mode & 1) return;                       /* skip initialisation */
    p->mode = mode;
    /* calculate max. delay time */
    npts = (int) (*p->imdel * ekr + FL(1.5));
    if (npts < 1) {
      initerror("vdelayk: invalid max delay time (must be >= 0)");
      return;
    }
    p->wrtp = 0; p->npts = npts;
    /* allocate space for delay buffer */
    if (p->aux.auxp == NULL ||
        (npts * (int) sizeof(MYFLT)) > p->aux.size) {
      auxalloc((long) (npts * sizeof(MYFLT)), &p->aux);
    }
    p->init_k = npts;           /* not -1 this time ! */
}

void vdelayk(VDELAYK *p)
{
    MYFLT   *buf = (MYFLT*) p->aux.auxp;
    int     n, npts = p->npts;

    if (!buf) {
      perferror("vdelayk: not initialised");
      return;
    }
    buf[p->wrtp] = *(p->ksig);              /* write input signal to buffer */
    n = (int) (*(p->kdel) * ekr + FL(0.5));     /* calculate delay time */
    if (n < 0) {
      perferror("vdelayk: invalid delay time (must be >= 0)");
      return;
    }
    n = p->wrtp - n;
    if (++p->wrtp >= npts) p->wrtp = 0;         /* wrap index */
    if (p->init_k) {
      if (p->mode & 2) {
        if (npts == p->init_k)
          p->frstkval = *(p->ksig);             /* save first input value */
        *(p->ar) = (n < 0 ? p->frstkval : buf[n]);      /* initial delay */
      }
      else {
        *(p->ar) = (n < 0 ? FL(0.0) : buf[n]);
      }
      p->init_k--;
    }
    else {
      while (n < 0) n += npts;
      *(p->ar) = buf[n];                        /* read output signal */
    }
}

/* ------------ rbjeq opcode ------------ */

/* original algorithm by Robert Bristow-Johnson */
/* Csound orchestra version by Josep M Comajuncosas, Aug 1999 */
/* ported to C (and optimised) by Istvan Varga, Dec 2002 */

/* ar rbjeq asig, kfco, klvl, kQ, kS[, imode] */

/* IV - Dec 28 2002: according to the original version by JMC, the formula */
/*   alpha = sin(omega) * sinh(1 / (2 * Q))                                */
/* should be used to calculate Q. However, according to my tests, it seems */
/* to be wrong with low Q values, where this simplified code               */
/*   alpha = sin(omega) / (2 * Q)                                          */
/* was measured to be more accurate. It also makes the Q value for no      */
/* resonance exactly sqrt(0.5) (as it would be expected), while the old    */
/* version required a Q setting of about 0.7593 for no resonance.          */
/* With Q >= 1, there is not much difference.                              */
/* N.B.: the above apply to the lowpass and highpass filters only. For     */
/* bandpass, band-reject, and peaking EQ, the modified formula is          */
/*   alpha = tan(omega / (2 * Q))                                          */

/* Defining this macro selects the revised version, while commenting it    */
/* out enables the original.                                               */

/* #undef IV_Q_CALC */
#define IV_Q_CALC 1

void rbjeqset(RBJEQ *p)
{
    int mode = (int) (*p->imode + FL(0.5)) & 0xF;

    if (mode & 1) return;               /* skip initialisation */
    /* filter type */
    p->ftype = mode >> 1;
    /* reset filter */
    p->old_kcps = p->old_klvl = p->old_kQ = p->old_kS = FL(-1.12123e35);
    p->b0 = p->b1 = p->b2 = p->a1 = p->a2 = FL(0.0);
    p->xnm1 = p->xnm2 = p->ynm1 = p->ynm2 = FL(0.0);
}

void rbjeq(RBJEQ *p)
{
    int     nn, new_frq;
    MYFLT   b0, b1, b2, a1, a2, tmp;
    MYFLT   xnm1, xnm2, ynm1, ynm2;
    MYFLT   *ar, *asig;
    double  dva0;

    if (*(p->kcps) != p->old_kcps) {
      /* frequency changed */
      new_frq = 1;
      p->old_kcps = *(p->kcps);
      /* calculate variables that depend on freq., and are used by all modes */
      p->omega = (double) p->old_kcps * TWOPI / (double) esr;
      p->cs = cos(p->omega);
      p->sn = sqrt(1.0 - p->cs * p->cs);
    }
    else
      new_frq = 0;
    /* copy object data to local variables */
    ar = p->ar; asig = p->asig;
    xnm1 = p->xnm1; xnm2 = p->xnm2; ynm1 = p->ynm1; ynm2 = p->ynm2;
    nn = (int) ksmps;
    switch (p->ftype) {
    case 0:                                     /* lowpass filter */
      if (new_frq || *(p->kQ) != p->old_kQ) {
        double  alpha;
        p->old_kQ = *(p->kQ);
#ifdef IV_Q_CALC
        alpha = p->sn * 0.5 / (double) p->old_kQ;       /* IV - Dec 28 2002 */
#else
        alpha = p->sn * sinh(0.5 / (double) p->old_kQ);
#endif
        /* recalculate all coeffs */
        dva0 = 1.0 / (1.0 + alpha);
        p->b2 = (MYFLT) (0.5 * (dva0 - dva0 * p->cs));
        p->a1 = (MYFLT) (-2.0 * dva0 * p->cs);
        p->a2 = (MYFLT) (dva0 - dva0 * alpha);
      }
      b2 = p->b2; a1 = p->a1; a2 = p->a2;
      do {
        tmp = *asig++;
        *ar = b2 * (tmp + xnm1 + xnm1 + xnm2) - a1 * ynm1 - a2 * ynm2;
        xnm2 = xnm1; xnm1 = tmp;
        ynm2 = ynm1; ynm1 = *ar++;
      } while (--nn);
      break;
    case 1:                                     /* highpass filter */
      if (new_frq || *(p->kQ) != p->old_kQ) {
        double  alpha;
        p->old_kQ = *(p->kQ);
#ifdef IV_Q_CALC
        alpha = p->sn * 0.5 / (double) p->old_kQ;       /* IV - Dec 28 2002 */
#else
        alpha = p->sn * sinh(0.5 / (double) p->old_kQ);
#endif
        /* recalculate all coeffs */
        dva0 = 1.0 / (1.0 + alpha);
        p->b2 = (MYFLT) (0.5 * (dva0 + dva0 * p->cs));
        p->a1 = (MYFLT) (-2.0 * dva0 * p->cs);
        p->a2 = (MYFLT) (dva0 - dva0 * alpha);
      }
      b2 = p->b2; a1 = p->a1; a2 = p->a2;
      do {
        tmp = *asig++;
        *ar = b2 * (tmp - xnm1 - xnm1 + xnm2) - a1 * ynm1 - a2 * ynm2;
        xnm2 = xnm1; xnm1 = tmp;
        ynm2 = ynm1; ynm1 = *ar++;
      } while (--nn);
      break;
    case 2:                                     /* bandpass filter */
      if (new_frq || *(p->kQ) != p->old_kQ) {
        double  alpha;
        p->old_kQ = *(p->kQ);
#ifdef IV_Q_CALC
        alpha = tan(p->omega * 0.5 / (double) p->old_kQ); /* IV - Dec 28 2002 */
#else
        alpha = p->sn * sinh(0.5 / (double) p->old_kQ);
#endif
        /* recalculate all coeffs */
        dva0 = 1.0 / (1.0 + alpha);
        p->b2 = (MYFLT) (dva0 * alpha);
        p->a1 = (MYFLT) (-2.0 * dva0 * p->cs);
        p->a2 = (MYFLT) (dva0 - dva0 * alpha);
      }
      b2 = p->b2; a1 = p->a1; a2 = p->a2;
      do {
        tmp = *asig++;
        *ar = b2 * (tmp - xnm2) - a1 * ynm1 - a2 * ynm2;
        xnm2 = xnm1; xnm1 = tmp;
        ynm2 = ynm1; ynm1 = *ar++;
      } while (--nn);
      break;
    case 3:                                     /* band-reject (notch) filter */
      if (new_frq || *(p->kQ) != p->old_kQ) {
        double  alpha;
        p->old_kQ = *(p->kQ);
#ifdef IV_Q_CALC
        alpha = tan(p->omega * 0.5 / (double) p->old_kQ); /* IV - Dec 28 2002 */
#else
        alpha = p->sn * sinh(0.5 / (double) p->old_kQ);
#endif
        /* recalculate all coeffs */
        dva0 = 1.0 / (1.0 + alpha);
        p->b2 = (MYFLT) dva0;
        p->a1 = (MYFLT) (-2.0 * dva0 * p->cs);
        p->a2 = (MYFLT) (dva0 - dva0 * alpha);
      }
      b2 = p->b2; a1 = p->a1; a2 = p->a2;
      do {
        tmp = *asig++;
        *ar = b2 * (tmp + xnm2) - a1 * (ynm1 - xnm1) - a2 * ynm2;
        xnm2 = xnm1; xnm1 = tmp;
        ynm2 = ynm1; ynm1 = *ar++;
      } while (--nn);
      break;
    case 4:                                     /* peaking EQ */
      if (new_frq || *(p->kQ) != p->old_kQ || *(p->klvl) != p->old_klvl) {
        double  sq, alpha, tmp1, tmp2;
        p->old_kQ = *(p->kQ);
        sq = sqrt((double) (p->old_klvl = *(p->klvl)));
#ifdef IV_Q_CALC
        alpha = tan(p->omega * 0.5 / (double) p->old_kQ); /* IV - Dec 28 2002 */
#else
        alpha = p->sn * sinh(0.5 / (double) p->old_kQ);
#endif
        /* recalculate all coeffs */
        tmp1 = alpha / sq;
        dva0 = 1.0 / (1.0 + tmp1);
        tmp2 = alpha * sq * dva0;
        p->b0 = (MYFLT) (dva0 + tmp2);
        p->b2 = (MYFLT) (dva0 - tmp2);
        p->a1 = (MYFLT) (-2.0 * dva0 * p->cs);
        p->a2 = (MYFLT) (dva0 - dva0 * tmp1);
      }
      b0 = p->b0; b2 = p->b2; a1 = p->a1; a2 = p->a2;
      do {
        tmp = *asig++;
        *ar = b0 * tmp + b2 * xnm2 - a1 * (ynm1 - xnm1) - a2 * ynm2;
        xnm2 = xnm1; xnm1 = tmp;
        ynm2 = ynm1; ynm1 = *ar++;
      } while (--nn);
      break;
    case 5:                                     /* low shelf */
      if (new_frq || *(p->klvl) != p->old_klvl || *(p->kS) != p->old_kS) {
        double sq, beta, tmp1, tmp2, tmp3, tmp4;
        sq = sqrt((double) (p->old_klvl = *(p->klvl)));
        p->old_kS = *(p->kS);
        beta = p->sn * sqrt(((double) p->old_klvl + 1.0) / p->old_kS
                            - (double) p->old_klvl + sq + sq - 1.0);
        /* recalculate all coeffs */
        tmp1 = sq + 1.0;
        tmp2 = sq - 1.0;
        tmp3 = tmp1 * p->cs;
        tmp4 = tmp2 * p->cs;
        dva0 = 1.0 / (tmp1 + tmp4 + beta);
        p->a1 = (MYFLT) (-2.0 * dva0 * (tmp2 + tmp3));
        p->a2 = (MYFLT) (dva0 * (tmp1 + tmp4 - beta));
        dva0 *= sq;
        p->b0 = (MYFLT) (dva0 * (tmp1 - tmp4 + beta));
        p->b1 = (MYFLT) ((dva0 + dva0) * (tmp2 - tmp3));
        p->b2 = (MYFLT) (dva0 * (tmp1 - tmp4 - beta));
      }
      b0 = p->b0; b1 = p->b1; b2 = p->b2; a1 = p->a1; a2 = p->a2;
      do {
        tmp = *asig++;
        *ar = b0 * tmp + b1 * xnm1 + b2 * xnm2 - a1 * ynm1 - a2 * ynm2;
        xnm2 = xnm1; xnm1 = tmp;
        ynm2 = ynm1; ynm1 = *ar++;
      } while (--nn);
      break;
    case 6:                                     /* high shelf */
      if (new_frq || *(p->klvl) != p->old_klvl || *(p->kS) != p->old_kS) {
        double sq, beta, tmp1, tmp2, tmp3, tmp4;
        sq = sqrt((double) (p->old_klvl = *(p->klvl)));
        p->old_kS = *(p->kS);
        beta = p->sn * sqrt(((double) p->old_klvl + 1.0) / p->old_kS
                            - (double) p->old_klvl + sq + sq - 1.0);
        /* recalculate all coeffs */
        tmp1 = sq + 1.0;
        tmp2 = sq - 1.0;
        tmp3 = tmp1 * p->cs;
        tmp4 = tmp2 * p->cs;
        dva0 = 1.0 / (tmp1 - tmp4 + beta);
        p->a1 = (MYFLT) ((dva0 + dva0) * (tmp2 - tmp3));
        p->a2 = (MYFLT) (dva0 * (tmp1 - tmp4 - beta));
        dva0 *= sq;
        p->b0 = (MYFLT) (dva0 * (tmp1 + tmp4 + beta));
        p->b1 = (MYFLT) (-2.0 * dva0 * (tmp2 + tmp3));
        p->b2 = (MYFLT) (dva0 * (tmp1 + tmp4 - beta));
      }
      b0 = p->b0; b1 = p->b1; b2 = p->b2; a1 = p->a1; a2 = p->a2;
      do {
        tmp = *asig++;
        *ar = b0 * tmp + b1 * xnm1 + b2 * xnm2 - a1 * ynm1 - a2 * ynm2;
        xnm2 = xnm1; xnm1 = tmp;
        ynm2 = ynm1; ynm1 = *ar++;
      } while (--nn);
      break;
    default:
      perferror("rbjeq: invalid filter type");
      return;
      break;
    }
    /* save filter state */
    p->xnm1 = xnm1; p->xnm2 = xnm2; p->ynm1 = ynm1; p->ynm2 = ynm2;
}

